from __future__ import annotations

import json
import os
import shutil
import subprocess
import tarfile
import tempfile
import uuid
from datetime import datetime, timezone
from functools import wraps
from itertools import chain, repeat
from pathlib import Path
from runpy import run_path
from typing import Any, Callable

import config
import distro
import requests
from requests.auth import HTTPBasicAuth

from eportal.backups import backup_databases
from eportal.patches_constants import Stream
from eportal.patchsource import PatchSource
from eportal.utils import get_instance_id, get_version

EXTERNAL_URLS = [
    'https://patches.kernelcare.com',
    'https://repo.cloudlinux.com/kernelcare/kernelcare_install.sh',
    'https://repo.eportal.kernelcare.com/',
    'https://downloads.kernelcare.com/',
]
GET_MY_IP_URL = 'https://patches.kernelcare.com/myip'
REPORT_UPLOAD_URI = 'upload/eportal-doctor'

EPORTAL_USER_CONFIG = '/etc/eportal/config'


def handle_exceptions(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return func(*args, **kwargs)
        except Exception as error:
            return {'error': repr(error)}

    return wrapper


def compact_patches_info(feeds: list[str], dashboard: dict[str, Any]) -> dict[str, Any]:
    status = dashboard['status']
    unpacked = dashboard['unpacked']
    patchsets = dashboard['patchsets']
    stitle = {None: 'Undeployed', True: 'Enabled', False: 'Disabled', 'NA': 'N/A'}

    ps_state: dict[str, dict[str, list[str]]] = {}
    for ps in patchsets:
        for f in feeds:
            key = f, ps
            if key in status:
                s = stitle.get(status[key], 'UNKNOWN')
            else:
                s = 'Undeployed' if ps in unpacked else 'Not downloaded'
            ps_state.setdefault(ps, {}).setdefault(s, []).append(f)

    return ps_state


@handle_exceptions
def get_patches_info() -> dict[str, dict[str, Any]]:
    # NOTE: eportal.web importing forces DB client initialization.
    # In case importing globally and DB files are corrupted it will fail to load.

    from eportal.patches_impl import get_repo
    from eportal.web.views.main import get_dashboard_content

    patches_info = {}
    dc = get_dashboard_content(0, None)
    for stream in Stream:
        repo = get_repo(0, stream)
        stat = repo.dashboard(dc.feeds, usage=False)
        patches_info[stream.value] = compact_patches_info(dc.feeds, stat)
    return patches_info


def get_proxy_settings() -> dict[str, str]:
    def get_proto_proxy(proto: str) -> str | None:
        return os.environ.get(proto.upper() + '_PROXY', os.environ.get(proto + '_proxy', config.PROXY))

    return {k: v for k, v in [(proto, get_proto_proxy(proto)) for proto in ('http', 'https')] if v}


def get_resource(url: str, auth: HTTPBasicAuth | None = None) -> tuple[int, dict, bytes]:
    s = requests.Session()
    s.proxies = get_proxy_settings()
    r = s.get(url, auth=auth, timeout=10)
    return r.status_code, dict(r.headers), r.content


def check_proxy() -> bool:
    for envvar in list(os.environ):
        if envvar.lower() in ('http_proxy', 'https_proxy'):
            return True
    return False


def check_connectivity() -> dict[str, dict[str, Any] | bool]:
    result: dict[str, dict[str, Any] | bool] = {'proxy': check_proxy()}

    ps = PatchSource()
    upload_url = [(ps.get_url(REPORT_UPLOAD_URI), HTTPBasicAuth(*ps.auth))] if ps.is_configured() else []
    for url, auth in chain(zip(EXTERNAL_URLS, repeat(None)), upload_url):
        try:
            status, headers, _ = get_resource(url, auth=auth)
            result[url] = {'status': status, 'headers': headers}
        except Exception as error:
            result[url] = {'error': repr(error)}
    return result


def run_command(command: str) -> tuple[str, str]:
    result = subprocess.run(command, shell=True, capture_output=True)  # nosec: B602
    out = result.stdout.decode('utf8', errors='replace')
    err = result.stderr.decode('utf8', errors='replace')
    return out, err


@handle_exceptions
def get_disk_usage() -> list[str]:
    output, _ = run_command('df -h')
    return output.split('\n')


@handle_exceptions
def get_os_version() -> dict[str, str]:
    distro_id = distro.id()
    distro_version = distro.version()
    distro_name = distro.name()
    return {'ID': distro_id, 'Version': distro_version, 'Name': distro_name}


@handle_exceptions
def get_process_list() -> list[str]:
    result, _ = run_command('ps aux | grep -E "(nginx|eportal)"')
    return result.split('\n')


def collect_logs(dest_dir: Path) -> None:
    for nginx_log in ('access.log', 'kcare-eportal.log', 'error.log'):
        run_command(f'tail -n 10000 /var/log/nginx/{nginx_log} > "{dest_dir / nginx_log}" 2>&1')
    run_command(f'journalctl -u eportal -ocat -n 10000 > "{dest_dir / "eportal.log"}" 2>&1')
    run_command(f'tail -n 10000 /var/log/audit/audit.log > "{dest_dir / "audit.log"}" 2>&1')


def is_sensitive_data_key(key: str) -> bool:
    return any(pattern in key for pattern in ('KEY', 'SECRET', 'DATABASE', 'PROXY', 'PASSWORD'))


def prepare_config(path: str) -> dict:
    config = run_path(path)
    hidden_value = '********'
    return {k: hidden_value if is_sensitive_data_key(k) else v for k, v in config.items() if k.isupper()}


@handle_exceptions
def get_config_options() -> dict[str, Any]:
    paths = [config.jroot('config', 'local.py'), config.jroot('config', 'docker.py'), EPORTAL_USER_CONFIG]
    eportal_config: dict[str, Any] = {}
    for path in paths:
        if os.path.exists(path):
            conf = prepare_config(path)
        else:
            conf = {'msg': 'No user defined config options.'}
        eportal_config[path] = conf
    return eportal_config


@handle_exceptions
def get_eportal_ip() -> str:
    _, _, content = get_resource(GET_MY_IP_URL)
    return content.decode('utf-8')


def get_eportal_database(dst: Path) -> None:
    try:
        backup_databases(None, str(dst))
    except Exception as error:
        (dst / 'error.log').write_text(f'Failed to copy DB: {error!r}')


def create_temp_dir() -> Path:
    while True:
        try:
            tmpdir = Path(tempfile.gettempdir()) / str(uuid.uuid4())
            tmpdir.mkdir(mode=0o700)
            return tmpdir
        except FileExistsError:  # pragma: no cover
            pass


def write_json(obj: Any, dest: Path) -> None:
    def encoder_stub(obj: Any) -> str:
        return f'Failed to encode object: {obj!r}'

    with dest.open('w') as json_file:
        json.dump(obj, json_file, ensure_ascii=False, indent=4, default=encoder_stub)


def create_archive(src_dir: Path) -> Path:
    for compression_mode in ('w:xz', 'w:bz2', 'w:gz'):
        dst = src_dir.with_name(f'{src_dir.name}.tar.{compression_mode[2:]}')
        try:
            with tarfile.open(str(dst), compression_mode) as tar:
                for entry in src_dir.iterdir():
                    tar.add(entry, arcname=entry.name)
            return dst
        except tarfile.CompressionError:
            dst.unlink(missing_ok=True)
    else:
        raise tarfile.CompressionError()


def make_doctor_report() -> Path:
    tmp_dir = create_temp_dir()
    try:
        version, platform = get_version()
        report = {
            'eportal_ip': get_eportal_ip(),
            'os_version': get_os_version(),
            'eportal_version': version,
            'platform': platform,
            'patches_info': get_patches_info(),
            'connectivity': check_connectivity(),
            'disk_usage': get_disk_usage(),
            'process_list': get_process_list(),
            'config': get_config_options(),
        }
        write_json(report, tmp_dir / 'report.json')

        logs_dir = tmp_dir / 'logs'
        logs_dir.mkdir(mode=0o700)
        collect_logs(logs_dir)

        db_dir = tmp_dir / 'db'
        db_dir.mkdir(mode=0o700)
        get_eportal_database(db_dir)

        manifest = {
            'schema_version': 1,
            'type': 'eportal-doctor',
            'time_created': datetime.now(tz=timezone.utc).isoformat(),
        }
        write_json(manifest, tmp_dir / 'manifest.json')

        return create_archive(tmp_dir)

    finally:
        shutil.rmtree(str(tmp_dir), ignore_errors=True)


def send_report(report_path: Path) -> None:
    ps = PatchSource()
    ps.check_configuration()

    with report_path.open('rb') as report_fileobj:
        response = requests.put(
            ps.get_url(f'{REPORT_UPLOAD_URI}/{report_path.name}'),
            auth=HTTPBasicAuth(*ps.auth),
            headers={'serverid': get_instance_id(), 'content-type': 'application/octet-stream'},
            data=report_fileobj,
            timeout=30,
            proxies=get_proxy_settings(),
        )
        response.raise_for_status()
